"""
Utility functions.

These will be migrated out into a base package at some point.
"""

from .intervals import Interval, IntervalTree, Moment
from .iterables import (
    expand,
    flatten_iterable,
    group_iterable_by_count,
    iterate_nwise,
    repeat_sequence_to_length,
    zip_sequences,
)

__all__ = [
    "Interval",
    "IntervalTree",
    "Moment",
    "expand",
    "flatten_iterable",
    "group_iterable_by_count",
    "iterate_nwise",
    "repeat_sequence_to_length",
    "zip_sequences",
]



from typing import Callable, Optional
from typing import Type as TypingType

from mypy.nodes import ARG_OPT, Argument, AssignmentStmt, CallExpr, RefExpr, Var
from mypy.plugin import ClassDefContext, Plugin
from mypy.plugins.common import _get_decorator_bool_argument
from mypy.typeops import make_simplified_union
from mypy.types import NoneType


class UGenTransformer:
    def __init__(self, ctx: ClassDefContext) -> None:
        self._ctx = ctx

    def collect_params(self):
        cls = self._ctx.cls
        params = []
        for stmt in cls.defs.body:
            if not isinstance(stmt, AssignmentStmt):
                continue
            expr = stmt.rvalue
            if (
                isinstance(expr, CallExpr)
                and isinstance(expr.callee, RefExpr)
                and expr.callee.fullname in ["supriya.ugens.bases.param"]
            ):
                params.append(stmt.lvalues[0].name)
        return params

    def transform(self) -> bool:
        # is_classmethod flag is not released yet
        from mypy.plugins.common import add_attribute_to_class, add_method_to_class

        api = self._ctx.api
        cls = self._ctx.cls
        info = self._ctx.cls.info

        ugen_input_type = make_simplified_union(
            [
                api.named_type("typing.SupportsFloat"),
                api.named_type("supriya.ugens.bases.UGenMethodMixin"),
            ]
        )

        decorator_arguments = {
            "ar": _get_decorator_bool_argument(self._ctx, "ar", False),
            "kr": _get_decorator_bool_argument(self._ctx, "kr", False),
            "ir": _get_decorator_bool_argument(self._ctx, "ir", False),
            "dr": _get_decorator_bool_argument(self._ctx, "dr", False),
            "new": _get_decorator_bool_argument(self._ctx, "new", False),
            "is_multichannel": _get_decorator_bool_argument(
                self._ctx, "is_multichannel", False
            ),
            "fixed_channel_count": _get_decorator_bool_argument(
                self._ctx, "fixed_channel_count", False
            ),
        }
        args = []
        for name in self.collect_params():
            args.append(
                Argument(
                    variable=Var(name, ugen_input_type),
                    type_annotation=ugen_input_type,
                    initializer=None,
                    kind=ARG_OPT,
                )
            )
            add_attribute_to_class(
                api=api,
                cls=cls,
                name=name,
                typ=ugen_input_type,
                override_allow_incompatible=True,
            )
        if (
            decorator_arguments["is_multichannel"]
            and not decorator_arguments["fixed_channel_count"]
        ):
            args.append(
                Argument(
                    variable=Var("channel_count", api.named_type("typing.SupportsInt")),
                    type_annotation=api.named_type("typing.SupportsInt"),
                    initializer=None,
                    kind=ARG_OPT,
                )
            )
        for name in ["ar", "kr", "ir", "dr", "new"]:
            if not decorator_arguments[name] or name in info.names:
                continue
            return_type = api.named_type("supriya.ugens.bases.UGenMethodMixin")
            if (
                decorator_arguments["is_multichannel"]
                or decorator_arguments["fixed_channel_count"]
            ):
                return_type = api.named_type("supriya.ugens.bases.UGenArray")
            add_method_to_class(
                api=api,
                cls=cls,
                name=name,
                args=args,
                return_type=return_type,
                is_classmethod=True,
            )
        if "__init__" not in info.names:
            add_method_to_class(
                api=api, cls=cls, name="__init__", args=args, return_type=NoneType()
            )
        return False


def _ugen_hook(ctx: ClassDefContext):
    UGenTransformer(ctx).transform()


class SupriyaPlugin(Plugin):
    def get_class_decorator_hook(
        self, fullname: str
    ) -> Optional[Callable[[ClassDefContext], None]]:
        if fullname == "supriya.ugens.bases.ugen":
            return _ugen_hook
        return None


def plugin(version: str) -> TypingType[SupriyaPlugin]:
    return SupriyaPlugin



from ...enums import EnvelopeShape
from ...synthdefs import Envelope, SynthDefBuilder
from ...ugens import EnvGen, In, InFeedback, Out

__all__ = []


def _build_link_audio_synthdef(channel_count):
    name = "system_link_audio_{}".format(channel_count)
    builder = SynthDefBuilder(
        name=name, out=0, in_=16, gate=1, fade_time=0.02, done_action=2
    )
    with builder:
        start_value = builder["fade_time"] <= 0
        envelope = Envelope(
            amplitudes=[start_value, 1.0, 0.0],
            durations=[1.0, 1.0],
            curves=[EnvelopeShape.SINE, -EnvelopeShape.SINE],
            release_node=1.0,
        )
        envelope = EnvGen.kr(
            done_action=builder["done_action"],
            envelope=envelope,
            gate=builder["gate"],
            time_scale=builder["fade_time"],
        )
        source = InFeedback.ar(bus=builder["in_"], channel_count=channel_count)
        Out.ar(bus=builder["out"], source=source * envelope)
    globals()[name] = builder.build()
    __all__.append(name)


def _build_link_control_synthdef(channel_count):
    name = "system_link_control_{}".format(channel_count)
    builder = SynthDefBuilder(
        name=name, out=0, in_=16, gate=1, fade_time=0.02, done_action=2
    )
    with builder:
        start_value = builder["fade_time"] <= 0
        envelope = Envelope(
            amplitudes=[start_value, 1.0, 0.0],
            durations=[1.0, 1.0],
            curves=EnvelopeShape.LINEAR,
            release_node=1.0,
        )
        envelope = EnvGen.kr(
            done_action=builder["done_action"],
            envelope=envelope,
            gate=builder["gate"],
            time_scale=builder["fade_time"],
        )
        input_ = In.kr(bus=builder["in_"], channel_count=channel_count)
        Out.kr(bus=builder["out"], source=input_ * envelope)
    globals()[name] = builder.build()
    __all__.append(name)


for i in range(1, 17):
    _build_link_audio_synthdef(i)
    _build_link_control_synthdef(i)



import asyncio
import logging
import random
import sys

import pytest

from supriya.contexts.realtime import (
    DEFAULT_HEALTHCHECK,
    AsyncServer,
    BootStatus,
    Server,
)
from supriya.exceptions import (
    OwnedServerShutdown,
    ServerCannotBoot,
    ServerOnline,
    TooManyClients,
    UnownedServerShutdown,
)
from supriya.osc import find_free_port

supernova = pytest.param(
    "supernova",
    marks=pytest.mark.skipif(
        sys.platform.startswith("win"), reason="Supernova won't boot on Windows"
    ),
)


async def get(x):
    if asyncio.iscoroutine(x):
        return await x
    return x


@pytest.fixture(autouse=True)
def use_caplog(caplog):
    caplog.set_level(logging.INFO)


@pytest.fixture(autouse=True)
def healthcheck_attempts(monkeypatch):
    monkeypatch.setattr(DEFAULT_HEALTHCHECK, "max_attempts", 1)


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_only(executable, context_class):
    context = context_class()
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    result = context.boot(executable=executable)
    if asyncio.iscoroutine(result):
        await result
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_and_quit(executable, context_class):
    context = context_class()
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    result = context.boot(executable=executable)
    if asyncio.iscoroutine(result):
        await result
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    result = context.quit()
    if asyncio.iscoroutine(result):
        await result
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_and_boot(executable, context_class):
    context = context_class()
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    result = context.boot(executable=executable)
    if asyncio.iscoroutine(result):
        await result
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    with pytest.raises(ServerOnline):
        result = context.boot(executable=executable)
        if asyncio.iscoroutine(result):
            await result
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_and_quit_and_quit(executable, context_class):
    context = context_class()
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    result = context.boot(executable=executable)
    if asyncio.iscoroutine(result):
        await result
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    result = context.quit()
    if asyncio.iscoroutine(result):
        await result
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    result = context.quit()
    if asyncio.iscoroutine(result):
        await result
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_and_connect(executable, context_class):
    context = context_class()
    assert context.boot_status == BootStatus.OFFLINE
    assert not context.is_owner
    result = context.boot(executable=executable)
    if asyncio.iscoroutine(result):
        await result
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner
    with pytest.raises(ServerOnline):
        result = context.connect()
        if asyncio.iscoroutine(result):
            await result
    assert context.boot_status == BootStatus.ONLINE
    assert context.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_a_and_boot_b_cannot_boot(executable, context_class):
    context_a, context_b = context_class(), context_class()
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    result = context_a.boot(maximum_logins=4, executable=executable)
    if asyncio.iscoroutine(result):
        await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    with pytest.raises(ServerCannotBoot):
        result = context_b.boot(maximum_logins=4, executable=executable)
        if asyncio.iscoroutine(result):
            await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner


# scsynth only
@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_a_and_connect_b_too_many_clients(executable, context_class):
    context_a, context_b = context_class(), context_class()
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    result = context_a.boot(maximum_logins=1)
    if asyncio.iscoroutine(result):
        await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    with pytest.raises(TooManyClients):
        result = context_b.connect()
        if asyncio.iscoroutine(result):
            await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_a_and_connect_b_and_quit_a(executable, context_class):
    context_a, context_b = context_class(), context_class()
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    result = context_a.boot(maximum_logins=2, executable=executable)
    if asyncio.iscoroutine(result):
        await result
    result = context_b.connect()
    if asyncio.iscoroutine(result):
        await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    result = context_a.quit()
    if asyncio.iscoroutine(result):
        await result
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    for _ in range(100):
        await asyncio.sleep(0.1)
        if context_b.boot_status == BootStatus.OFFLINE:
            break
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_a_and_connect_b_and_disconnect_b(executable, context_class):
    context_a, context_b = context_class(), context_class()
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    result = context_a.boot(maximum_logins=2, executable=executable)
    if asyncio.iscoroutine(result):
        await result
    result = context_b.connect()
    if asyncio.iscoroutine(result):
        await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    result = context_b.disconnect()
    if asyncio.iscoroutine(result):
        await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_a_and_connect_b_and_disconnect_a(executable, context_class):
    context_a, context_b = context_class(), context_class()
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    result = context_a.boot(maximum_logins=2, executable=executable)
    if asyncio.iscoroutine(result):
        await result
    result = context_b.connect()
    if asyncio.iscoroutine(result):
        await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    with pytest.raises(OwnedServerShutdown):
        result = context_a.disconnect()
        if asyncio.iscoroutine(result):
            await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_a_and_connect_b_and_quit_b(executable, context_class):
    context_a, context_b = context_class(), context_class()
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    result = context_a.boot(maximum_logins=2, executable=executable)
    if asyncio.iscoroutine(result):
        await result
    result = context_b.connect()
    if asyncio.iscoroutine(result):
        await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    with pytest.raises(UnownedServerShutdown):
        result = context_b.quit()
        if asyncio.iscoroutine(result):
            await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_a_and_connect_b_and_force_quit_b(executable, context_class):
    context_a, context_b = context_class(), context_class()
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    result = context_a.boot(maximum_logins=2, executable=executable)
    if asyncio.iscoroutine(result):
        await result
    result = context_b.connect()
    if asyncio.iscoroutine(result):
        await result
    assert context_a.boot_status == BootStatus.ONLINE and context_a.is_owner
    assert context_b.boot_status == BootStatus.ONLINE and not context_b.is_owner
    result = context_b.quit(force=True)
    if asyncio.iscoroutine(result):
        await result
    assert context_b.boot_status == BootStatus.OFFLINE and not context_b.is_owner
    for _ in range(100):
        await asyncio.sleep(0.1)
        if context_a.boot_status == BootStatus.OFFLINE:
            break
    assert context_a.boot_status == BootStatus.OFFLINE and not context_a.is_owner


@pytest.mark.asyncio
@pytest.mark.parametrize("executable", ["scsynth", supernova])
@pytest.mark.parametrize("context_class", [AsyncServer, Server])
async def test_boot_reboot_sticky_options(executable, context_class):
    """
    Options persist across booting and quitting.
    """
    context = context_class()
    maximum_node_count = random.randint(1024, 2048)
    await get(
        context.boot(maximum_node_count=maximum_node_count, port=find_free_port())
    )
    assert context.options.maximum_node_count == maximum_node_count
    await get(context.quit())
    assert context.options.maximum_node_count == maximum_node_count
    await get(context.boot(port=find_free_port()))
    assert context.options.maximum_node_count == maximum_node_count
    await get(context.quit())
    assert context.options.maximum_node_count == maximum_node_count



import operator

import pytest

from supriya.patterns import BinaryOpPattern, SequencePattern, UnaryOpPattern


@pytest.mark.parametrize(
    "op, expr_one, expr_two, expected",
    [
        (
            operator.add,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern("+", 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.add,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern("+", SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.floordiv,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern("//", 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.floordiv,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern("//", SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.mod,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern("%", 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.mod,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern("%", SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.mul,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern("*", 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.mul,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern("*", SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.pow,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern("**", 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.pow,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern("**", SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.sub,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern("-", 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.sub,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern("-", SequencePattern([1, 2, 3]), 23),
        ),
        (
            operator.truediv,
            23,
            SequencePattern([1, 2, 3]),
            BinaryOpPattern("/", 23, SequencePattern([1, 2, 3])),
        ),
        (
            operator.truediv,
            SequencePattern([1, 2, 3]),
            23,
            BinaryOpPattern("/", SequencePattern([1, 2, 3]), 23),
        ),
    ],
)
def test_binary_ops(op, expr_one, expr_two, expected):
    if isinstance(expected, Exception):
        with pytest.raises(expected):
            op(expr_one, expr_two)
    else:
        assert op(expr_one, expr_two) == expected


@pytest.mark.parametrize(
    "op, expr, expected",
    [
        (
            operator.abs,
            SequencePattern([1, 2, 3]),
            UnaryOpPattern("abs", SequencePattern([1, 2, 3])),
        ),
        (
            operator.inv,
            SequencePattern([1, 2, 3]),
            UnaryOpPattern("~", SequencePattern([1, 2, 3])),
        ),
        (
            operator.neg,
            SequencePattern([1, 2, 3]),
            UnaryOpPattern("-", SequencePattern([1, 2, 3])),
        ),
        (
            operator.pos,
            SequencePattern([1, 2, 3]),
            UnaryOpPattern("+", SequencePattern([1, 2, 3])),
        ),
    ],
)
def test_unary_ops(op, expr, expected):
    if isinstance(expected, Exception):
        with pytest.raises(expected):
            op(expr)
    else:
        assert op(expr) == expected



from ...enums import ParameterRate
from ...synthdefs import Parameter, SynthDefBuilder
from ...ugens import Out, SinOsc


def _build_test_synthdef():
    with SynthDefBuilder(
        frequency=440,
        amplitude=Parameter(value=1.0, parameter_rate=ParameterRate.AUDIO),
    ) as builder:
        sin_osc = SinOsc.ar(frequency=builder["frequency"])
        enveloped_sin = sin_osc * builder["amplitude"]
        Out.ar(bus=0, source=enveloped_sin)
    synthdef = builder.build(name="test")
    return synthdef


test = _build_test_synthdef()

__all__ = ("test",)
